Imports RegExp = System.Text.RegularExpressions
Imports madrat.Calculate.Expressions
Imports madrat.Calculate.Parser
Imports madrat.Utility

Namespace Calculate

    ''' <completionlist cref="madrat.Calculate.Command"></completionlist>
    Partial Public NotInheritable Class Command

        <Flags()> _
        Private Enum InspectWhat
            Operators = 1
            Functions = 2
            UserDef = 4
            Variables = 8
            All = Operators Or Functions Or UserDef Or Variables
        End Enum

        Private Shared Sub DoInspection(ByVal cmd As String)
            Dim scanner As New WordScanner(cmd)
            Dim what As InspectWhat

            scanner.Match(Of CommandToken)()

            If scanner.EndOfStream Then
                what = InspectWhat.All
            Else
                If TypeOf scanner.Lookahead Is IdentifierToken Then
                    Select Case DirectCast(scanner.Lookahead, IdentifierToken).Identifier.ToLower()
                        Case "all", "a" : what = InspectWhat.All
                        Case "operators", "o" : what = InspectWhat.Operators
                        Case "functions", "f" : what = InspectWhat.Functions
                        Case "user", "u" : what = InspectWhat.UserDef
                        Case "variables", "v" : what = InspectWhat.Variables
                        Case Else
                            Display.WrapTextOut( _
                                "You have entered no valid parameter. Type ':help inspect' for more information." _
                            )
                            Return
                    End Select
                Else
                    Display.WrapTextOut( _
                        "You have entered no valid parameter. Type ':help inspect' for more information." _
                    )
                    Return
                End If
            End If

            If (what And InspectWhat.Operators) = InspectWhat.Operators Then
                Dim t As Type = GetType(Operation)

                For Each m As Reflection.MemberInfo In t.FindMembers( _
                    Reflection.MemberTypes.Field, _
                    Reflection.BindingFlags.Public Or _
                    Reflection.BindingFlags.Static, _
                    Nothing, Nothing _
                )
                    Dim o As Operation = DirectCast(t.InvokeMember( _
                        m.Name, _
                        Reflection.BindingFlags.GetField Or _
                        Reflection.BindingFlags.Public Or _
                        Reflection.BindingFlags.Static, _
                        Nothing, Nothing, Nothing _
                    ), Operation)
                    Console.Write(o)
                    Console.Write(ControlChars.Tab)
                Next

                Console.WriteLine()
            End If

            If (what And InspectWhat.Functions) = InspectWhat.Functions Then
                Static funs As String() = New String() { _
                    "sin", "cos", "tan", "exp", "log", "ln", "lg", "sqrt", "fact", "ans" _
                }

                For Each f As String In funs
                    Console.Write(f)
                    Console.Write(ControlChars.Tab)
                Next

                Console.WriteLine()
            End If

            If (what And InspectWhat.UserDef) = InspectWhat.UserDef Then
                Dim it As Dictionary(Of String, [Function]).KeyCollection.Enumerator = Functions.Iterator()

                Do While it.MoveNext()
                    Console.Write(it.Current)
                    Console.Write(ControlChars.Tab)
                Loop

                Console.WriteLine()
            End If

            If (what And InspectWhat.Variables) = InspectWhat.Variables Then
                Dim it As Dictionary(Of String, Values).KeyCollection.Enumerator = Variables.Iterator()

                Do While it.MoveNext()
                    Console.Write(it.Current)
                    Console.Write(ControlChars.Tab)
                Loop

                Console.WriteLine()
            End If
        End Sub

        Private Shared Sub DoClear(ByVal cmd As String)
            Console.Clear()
        End Sub

        Private Shared Sub ChangeMode(ByVal cmd As String)
            Dim scanner As New WordScanner(cmd)

            scanner.Match(Of CommandToken)()

            If TypeOf scanner.Lookahead Is IdentifierToken Then
                Select Case DirectCast(scanner.Lookahead, IdentifierToken).Identifier.ToLower()
                    Case "infix", "i"
                        Reader.ExpressionFormat = ExpressionFormat.Infix
                    Case "prefix", "r"
                        Reader.ExpressionFormat = ExpressionFormat.Prefix
                    Case "postfix", "o"
                        Reader.ExpressionFormat = ExpressionFormat.Postfix
                    Case Else
                        Display.WrapTextOut( _
                            "You have entered no valid new mode. Type ':help mode' for more information." _
                        )
                End Select
            Else
                Display.WrapTextOut("You have entered no valid new mode. Type ':help mode' for more information.")
            End If
        End Sub

        Private Shared Sub ChangePrecision(ByVal cmd As String)
            Dim scanner As New WordScanner(cmd)

            scanner.Match(Of CommandToken)()

            If TypeOf scanner.Lookahead Is IdentifierToken Then
                Select Case DirectCast(scanner.Lookahead, IdentifierToken).Identifier.ToLower()
                    Case "high", "h"
                        Reader.Precision = InputReader.DisplayPrecision.High
                    Case "low", "l"
                        Reader.Precision = InputReader.DisplayPrecision.Low
                    Case Else
                        Display.WrapTextOut( _
                            "You have entered no valid new precision. Type ':help prec' for more information." _
                        )
                End Select
            Else
                Display.WrapTextOut("You have entered no valid new precision. Type ':help prec' for more information.")
            End If
        End Sub

        Private Shared Sub AssignVariable(ByVal cmd As String)
            Const pattern As String = "(?<Command>^:[a-z]+)\s+(?<Variable>\$[a-z][a-z0-9]*)\s*=\s*(?<Expression>.+)"
            Static search As New RegExp.Regex( _
                pattern, _
                RegExp.RegexOptions.Compiled Or _
                RegExp.RegexOptions.ExplicitCapture Or _
                RegExp.RegexOptions.IgnoreCase _
            )

            Dim match As RegExp.Match = search.Match(cmd)

            If Not match.Success Then
                Display.WrapTextOut("The input was invalid. Type ':help assign' for further information.")
                Return
            End If

            Dim var As String = match.Groups("Variable").Value

            Variables.Item(var) = Reader.CalculateValues(cmd, match.Groups("Expression").Index, Reader.ExpressionFormat)
        End Sub

        Private Shared Sub DefineFunction(ByVal cmd As String)
            Dim scanner As New InfixScanner(cmd)

            scanner.Match(Of CommandToken)()

            Try
                If Not TypeOf scanner.Lookahead Is Parser.FunctionToken Then _
                    Throw New ParseException(0, "Function name missing.")

                Dim fun As String = DirectCast(scanner.Lookahead, Parser.FunctionToken).Function

                scanner.Match(Of Parser.FunctionToken)()
                If Not scanner.Match(Of OpenBraceToken)() Then Throw New ParseException(0, "Opening bracket missing.")

                Dim params As New List(Of String)()

                If Not TypeOf scanner.Lookahead Is Parser.VariableToken Then _
                    Throw New ParseException(0, "At least one parameter needed.")

                params.Add(DirectCast(scanner.Lookahead, Parser.VariableToken).Variable)
                scanner.Match(Of Parser.VariableToken)()

                Do While TypeOf scanner.Lookahead Is CommaToken
                    scanner.Match(Of CommaToken)()

                    If Not TypeOf scanner.Lookahead Is Parser.VariableToken Then _
                        Throw New ParseException(0, "Unexpected comma or missing parameter.")

                    params.Add(DirectCast(scanner.Lookahead, Parser.VariableToken).Variable)
                    scanner.Match(Of Parser.VariableToken)()
                Loop

                If Not scanner.Match(Of CloseBraceToken)() Then Throw New ParseException(0, "Closing bracket missing.")
                If Not scanner.Match(Of AssignToken)() Then Throw New ParseException(0, "The input was invalid.")

                Dim exp As ParsedExpression = Expression.Parse(cmd, scanner.Position, Reader.ExpressionFormat)

                Functions.Add(fun, params.ToArray(), exp)
            Catch ex As ParseException
                If ex.position = 0 Then
                    Display.WrapTextOut(ex.Message & " Type ':help def' for further information.")
                Else
                    Display.WrapTextOut(ex.ToString())
                End If
            Catch ex As ArgumentException
                Display.WrapTextOut( _
                    "The function does already exist. If you want to overwrite it, you have to delete it first." _
                )
            End Try
        End Sub

        Private Shared Sub UnDefine(ByVal cmd As String)
            Dim scanner As New InfixScanner(cmd)

            scanner.Match(Of CommandToken)()

            Try
                Do
                    Select Case True
                        Case TypeOf scanner.Lookahead Is Parser.VariableToken
                            Dim var As String = DirectCast(scanner.Lookahead, Parser.VariableToken).Variable

                            If Not Variables.Contains(var) Then
                                Display.WrapTextOut("Variable {0} is not defined.", var)
                            Else
                                Variables.Remove(var)
                            End If
                        Case TypeOf scanner.Lookahead Is Parser.FunctionToken
                            Dim fun As String = DirectCast(scanner.Lookahead, Parser.FunctionToken).Function

                            If Not Functions.Contains(fun) Then
                                Display.WrapTextOut("Function '{0}' is not defined.", fun)
                            Else
                                Functions.Remove(fun)
                            End If
                        Case Else
                            Throw New ParseException(0)
                    End Select

                    scanner.Match(scanner.Lookahead)

                    Select Case True
                        Case TypeOf scanner.Lookahead Is EndOfStreamToken
                            Exit Do
                        Case TypeOf scanner.Lookahead Is CommaToken
                            scanner.Match(Of CommaToken)()
                        Case Else
                            Throw New ParseException(0)
                    End Select
                Loop
            Catch ex As ParseException
                Display.WrapTextOut("The input was invalid. Type ':help undef' for further information.")
            End Try
        End Sub

        Private Shared Sub DisplayHelp(ByVal cmd As String)
            Dim scanner As New WordScanner(cmd)

            scanner.Match(Of CommandToken)()

            If Not scanner.EndOfStream Then
                Try
                    Command.Parse(DirectCast(scanner.Lookahead, Parser.IdentifierToken).Identifier).DisplayHelp()
                Catch ex As Exception
                    Display.WrapTextOut("The command you entered does not exist.")
                End Try
            Else
                Dim op As New IO.StringWriter()

                Const example_infix As String = "(2 ** 3 + 6) / 7"
                Const example_postfix As String = "2 3 ** 6 + 7 /"
                Const example_prefix As String = "/ + ** 2 3 6 7"
                Dim example As String = Nothing

                Select Case Reader.ExpressionFormat
                    Case Expressions.ExpressionFormat.Infix
                        example = example_infix
                    Case Expressions.ExpressionFormat.Postfix
                        example = example_postfix
                    Case Expressions.ExpressionFormat.Prefix
                        example = example_prefix
                End Select

                op.WriteLine("Enter any valid expression in the command line and hit <Return> to solve it.")
                op.WriteLine()
                op.WriteLine("Example:")
                op.WriteLine()
                op.WriteLine("    > {0}", example)
                op.WriteLine("    2")
                op.WriteLine()
                op.WriteLine("The application supports lists of expressions. There are two types of lists:")
                op.WriteLine("    - Semicolon-separated lists, where expressions are separated by semicola.")
                op.WriteLine("    - Ranges, specified in the format <from> .. <to> [: <step>].")
                op.WriteLine("      A semicolon-separated list may contain ranges.")
                op.WriteLine()
                op.WriteLine("        from     The start value of the range.")
                op.WriteLine("        to       The end value of the range.")
                op.WriteLine("        step     The increment (or decrement) step.")
                op.WriteLine("                 Optional (default value is 1).")
                op.WriteLine()
                op.WriteLine("Lists of values are only supported in infix expressions.")
                op.WriteLine()
                op.WriteLine("Example:")
                op.WriteLine()
                op.WriteLine("    > 2 .. 1 : -0.2; exp(1); ln(e)")
                op.WriteLine("    2")
                op.WriteLine("    1.8")
                op.WriteLine("    1.6")
                op.WriteLine("    1.4")
                op.WriteLine("    1.2")
                op.WriteLine("    1")
                op.WriteLine("    2.78128182845905")
                op.WriteLine("    1")
                op.WriteLine()
                op.WriteLine("Commands are preceded by a colon ("":""). Valid commands are:")
                op.WriteLine()

                Dim t As Type = GetType(Command)

                For Each m As Reflection.MemberInfo In t.FindMembers( _
                    Reflection.MemberTypes.Field, _
                    Reflection.BindingFlags.Public Or _
                    Reflection.BindingFlags.Static, _
                    Nothing, Nothing _
                )
                    Dim c As Command = DirectCast(t.InvokeMember( _
                        m.Name, _
                        Reflection.BindingFlags.GetField Or _
                        Reflection.BindingFlags.Public Or _
                        Reflection.BindingFlags.Static, _
                        Nothing, Nothing, Nothing _
                    ), Command)
                    op.WriteLine( _
                        "    {0,-10} {1,-2} {2}", c.Name, c.ShortName, _
                        Utility.Display.WrapLongText(c.Description, 79, 18).Trim() _
                    )
                Next

                op.WriteLine()
                op.WriteLine("Enter ':help <cmd>' to get more information on a specific command.")
                Utility.Display.LongTextOut(op)
            End If
        End Sub

        Private Shared Sub DisplayInfo(ByVal cmd As String)
            Console.WriteLine("{0} v{1}", My.Application.Info.ProductName, My.Application.Info.Version)
            Console.WriteLine(My.Application.Info.Copyright)
        End Sub

    End Class

End Namespace