Namespace Calculate

    Public NotInheritable Class Command

#Region " Public "

        Public ReadOnly Name As String
        Public ReadOnly ShortName As String
        Public ReadOnly Description As String
        Public ReadOnly Syntax As String
        Public ReadOnly HelpText As String

        Public Delegate Sub ExecutionHandler(ByVal cmd As String)

#Region " Properties "

        Public ReadOnly Property Parameters() As List(Of Parameter)
            Get
                Return Me._parameters
            End Get
        End Property

#End Region

#Region " Methods "

        Public Overrides Function ToString() As String
            Return Name
        End Function

        Public Overrides Function Equals(ByVal obj As Object) As Boolean
            Return Me = TryCast(obj, Command)
        End Function

        Public Overrides Function GetHashCode() As Integer
            Return Name.GetHashCode()
        End Function

        Public Function ToStringBuilder() As Text.StringBuilder
            Dim ret As New IO.StringWriter()

            If ShortName.Length > 0 Then ret.WriteLine(":{0}, :{1}", Name, ShortName)
            ret.WriteLine()
            ret.Write(Utility.Display.WrapLongText(Description, 79))

            If Parameters.Count > 0 Then
                ret.WriteLine()
                ret.WriteLine("Syntax:")
                ret.WriteLine()
                ret.Write("    ")
                ret.WriteLine(Syntax)
                ret.WriteLine()

                For Each param As Parameter In Parameters
                    ret.Write(param)
                Next
            End If

            If HelpText.Length > 0 Then
                ret.WriteLine()
                ret.Write(Utility.Display.WrapLongText(HelpText, 79))
            End If

            Return ret.GetStringBuilder()
        End Function

        Public Sub DisplayHelp()
            Utility.Display.LongTextOut(New IO.StringWriter(ToStringBuilder()))
        End Sub

        Public Function IsKnownAs(ByVal id As String) As Boolean
            Return String.Compare(Name, id, True) = 0 OrElse String.Compare(ShortName, id, True) = 0
        End Function

        Public Sub Execute(ByVal cmd As String)
            If ExecHandler IsNot Nothing Then ExecHandler(cmd)
        End Sub

#End Region

#Region " Static "

        Public Shared ReadOnly Quit As New Command("quit", "q", "Quits the calculator.", "", "", Nothing)
        Public Shared ReadOnly Info As New Command("info", "", "Prints the info message.", "", "", AddressOf DisplayInfo)
        Public Shared ReadOnly Help As New Command( _
            "help", "h", "Displays the help message.", _
            ":help [<cmd>]", "If no <cmd> was specified, a general help message is displayed.", _
            AddressOf DisplayHelp _
        )
        Public Shared ReadOnly Clear As New Command( _
            "clear", "c", "Clears the screen.", _
            ":clear", "", AddressOf DoClear _
        )
        ' TODO: Implement 'Save'.
        Public Shared ReadOnly Save As New Command( _
            "save", "s", "Saves user variables and functions to a file.", _
            ":save [(<fun1> [, <fun2> ...])] [in <file>]", "", Nothing _
        )
        ' TODO: Implement 'Load'.
        Public Shared ReadOnly Load As New Command( _
            "load", "l", "Loads user variables and functions from a file.", _
            ":load [(<fun1> [, <fun2> ...])] [from <file>]", "", Nothing _
        )
        Public Shared ReadOnly Inspect As New Command( _
            "inspect", "i", "Displays the available variables, constants, functions and operators.", _
            ":inspect [(all, functions, operators, user, variables)]", "", AddressOf DoInspection _
        )
        Public Shared ReadOnly Assign As New Command( _
            "assign", "a", "Assigns a value to a variable.", _
            ":assign <var> = <expression>", _
            "Example:" & ControlChars.NewLine & ControlChars.NewLine & _
            "    > :assign $p2 = 2 * p" & ControlChars.NewLine & _
            "    6.28318530717959", _
            AddressOf AssignVariable _
        )
        Public Shared ReadOnly Def As New Command( _
            "def", "d", "Defines a function.", _
            ":def <f>(<x> [, <y> ...]) = <expression>", _
            "The variables names must follow the convention for variables. For further information read the help for ':assign'." & _
            ControlChars.NewLine & ControlChars.NewLine & _
            "Example:" & ControlChars.NewLine & ControlChars.NewLine & "    > :def log2($x) =lg($x, 2)", _
            AddressOf DefineFunction _
        )
        Public Shared ReadOnly UnDef As New Command( _
            "undef", "u", "Deletes a user-defined function or variable.", _
            ":undef <a> [, <b> ...]", "", AddressOf UnDefine _
        )
        Public Shared ReadOnly Mode As New Command( _
            "mode", "m", "Changes the mode for expressions.", _
            ":mode infix|prefix|postfix", "", AddressOf ChangeMode _
        )
        Public Shared ReadOnly Prec As New Command( _
            "prec", "p", "Changes the display precision of the result values.", _
            ":prec high|low", _
            "Notice that only the display precision will be changed; internally, all values are handled " & _
            "using double precision.", AddressOf ChangePrecision _
        )

        Shared Sub New()
            Help.Parameters.Add(New Parameter("cmd", "Optional. Displays the specific help message for <cmd>."))

            Inspect.Parameters.Add(New Parameter("all", "a", "Display all available predefined functions and operators and all user defined functions and variables (default)."))
            Inspect.Parameters.Add(New Parameter("functions", "f", "Displays only predefined functions."))
            Inspect.Parameters.Add(New Parameter("operators", "o", "Displays only predefined operators."))
            Inspect.Parameters.Add(New Parameter("user", "u", "Displays user defined functions."))
            Inspect.Parameters.Add(New Parameter("variables", "v", "Displays all variables."))

            Assign.Parameters.Add(New Parameter("var", "The variable name consisting of the character '$', an alphabetic character and any number of alphabetic characters or numbers (e.g. '$n', '$x1', '$pi2'). If the name already exists, it will be overwritten."))
            Assign.Parameters.Add(New Parameter("expression", "The assigned value. This may be any valid expression or a list of expressions."))

            Def.Parameters.Add(New Parameter("f", "The name of the function, which must consist of two or three letters or digits, preceeded by a letter. The name must be unique."))
            Def.Parameters.Add(New Parameter("x, y, ...", "The function parameters."))
            Def.Parameters.Add(New Parameter("expression", "The assigned expression."))

            UnDef.Parameters.Add(New Parameter("a, b, ...", "The functions or variables to delete."))

            Mode.Parameters.Add(New Parameter("infix", "i", "Sets the expression syntax to infix."))
            Mode.Parameters.Add(New Parameter("prefix", "r", "Sets the expression syntax to prefix."))
            Mode.Parameters.Add(New Parameter("postfix", "o", "Sets the expression syntax to postfix."))

            Prec.Parameters.Add(New Parameter("high", "h", "Sets a high precision (up to 15 decimal places)."))
            Prec.Parameters.Add(New Parameter("low", "l", "Sets a low precision (up to 5 decimal places)."))
        End Sub

#Region " Methods "

        <DebuggerHidden()> _
        Public Shared Operator =(ByVal lhs As Command, ByVal rhs As Command) As Boolean
            Return lhs Is rhs
        End Operator

        <DebuggerHidden()> _
        Public Shared Operator <>(ByVal lhs As Command, ByVal rhs As Command) As Boolean
            Return lhs IsNot rhs
        End Operator

        ''' <summary>
        ''' Parses a command.
        ''' </summary>
        ''' <param name="cmd">The command to parse.</param>
        ''' <exception cref="Expressions.ParseException">Thrown if the expression is not valid.</exception>
        ''' <returns>Returns the parsed command.</returns>
        Public Shared Function Parse(ByVal cmd As String) As Command
            Dim ret As Command = Nothing
            If Not TryParse(cmd, ret) Then Throw New Expressions.ParseException(0)
            Return ret
        End Function

        ''' <summary>
        ''' Tries to parse a command.
        ''' </summary>
        ''' <param name="cmd">The command to parse.</param>
        ''' <param name="ret">The parsed command.</param>
        ''' <returns>Returns whether parsing was successful.</returns>
        Public Shared Function TryParse(ByVal cmd As String, ByRef ret As Command) As Boolean
            Dim t As Type = GetType(Command)

            If cmd(0) = ":"c Then cmd = cmd.Substring(1)

            For Each m As Reflection.MemberInfo In t.FindMembers( _
                Reflection.MemberTypes.Field, _
                Reflection.BindingFlags.Public Or _
                Reflection.BindingFlags.Static, _
                Nothing, Nothing _
            )
                Dim c As Command = DirectCast(t.InvokeMember( _
                    m.Name, _
                    Reflection.BindingFlags.GetField Or _
                    Reflection.BindingFlags.Public Or _
                    Reflection.BindingFlags.Static, _
                    Nothing, Nothing, Nothing _
                ), Command)
                If c.IsKnownAs(cmd) Then
                    ret = c
                    Return True
                End If
            Next

            Return False
        End Function

#End Region

#End Region

#End Region

#Region " Private "

        Private ReadOnly ExecHandler As ExecutionHandler
        Private _parameters As New List(Of Parameter)

#Region " Constructors "

        ''' <remarks>As this is an enumeration type, external creation of new objects is forbidden.</remarks>
        Private Sub New( _
            ByVal name As String, _
            ByVal short_name As String, _
            ByVal description As String, _
            ByVal syntax As String, _
            ByVal help_text As String, _
            ByVal exec_handler As ExecutionHandler _
        )
            Me.Name = name
            Me.ShortName = short_name
            Me.Description = description
            Me.Syntax = syntax
            Me.HelpText = help_text
            Me.ExecHandler = exec_handler
        End Sub

#End Region

#End Region

    End Class

End Namespace