Imports madrat.Calculate.Expressions

Namespace Calculate

    ''' <completionlist cref="madrat.Calculate.Operation"></completionlist>
    Public NotInheritable Class Operation

#Region " Public "

        ''' <summary>
        ''' Returns whether the operation takes two operands instead of only one.
        ''' </summary>
        Public ReadOnly NumberOfOperands As Integer

#Region " Methods "

        Public Overrides Function ToString() As String
            Return Representation.Trim()
        End Function

        Public Overrides Function Equals(ByVal obj As Object) As Boolean
            Return Me = TryCast(obj, Operation)
        End Function

        Public Overrides Function GetHashCode() As Integer
            Return Representation.GetHashCode()
        End Function

        ''' <summary>
        ''' Performs the arithmetical operation on one or more operands.
        ''' </summary>
        ''' <param name="values">The operands to perform on.</param>
        ''' <returns>Returns the result of the operation.</returns>
        Public Function Perform(ByVal ParamArray values As Values()) As Values
            Select Case Me
                Case Add
                    Return values(0) + values(1)
                Case Subtract
                    Return values(0) - values(1)
                Case BitwiseOr
                    Return values(0) Or values(1)
                Case BitwiseXor
                    Return values(0) Xor values(1)
                Case Multipy
                    Return values(0) * values(1)
                Case Divide
                    Return values(0) / values(1)
                Case IntegerDivide
                    Return values(0) \ values(1)
                Case Modulus
                    Return values(0) Mod values(1)
                Case BitwiseAnd
                    Return values(0) And values(1)
                Case Power
                    Return values(0) ^ values(1)
                Case Root
                    Return values(0) ^ (1 / values(1))
                Case BitShiftLeft
                    ' HACK: The following does not work because the second parameter of '<<' MUST be of type Integer.
                    'Return values(0) << values(1)
                    Return Expressions.Values.op_LeftShift(values(0), values(1))
                Case BitShiftRight
                    ' HACK: The following does not work because the second parameter of '>>' MUST be of type Integer.
                    'Return values(0) >> values(1)
                    Return Expressions.Values.op_RightShift(values(0), values(1))
                Case PositiveSign
                    Return values(0)
                Case NegativeSign
                    Return -values(0)
                Case Negate
                    Return Not values(0)
                Case DotDot
                    Dim from As Double = values(0)(0)
                    Dim [to] As Double = values(1)(0)
                    Dim [step] As Double = values(2)(0)
                    Dim ret As New Values()

                    For d As Double = from To [to] Step [step]
                        ret.Add(d)
                    Next

                    Return ret
                Case Colon
                    ' Not a regular operator, should not occur.
                    Return Nothing
                Case Semicolon
                    Dim ret As New Values(values(0))

                    ret.AddRange(values(1))
                    Return ret
            End Select

            Return New Values()
        End Function

#End Region

#Region " Static "

        Public Shared ReadOnly Add As New Operation("+", 2)
        Public Shared ReadOnly Subtract As New Operation("-", 2)
        Public Shared ReadOnly BitwiseOr As New Operation("|", 2)
        Public Shared ReadOnly BitwiseXor As New Operation("^", 2)

        Public Shared ReadOnly Multipy As New Operation("*", 2)
        Public Shared ReadOnly Divide As New Operation("/", 2)
        Public Shared ReadOnly IntegerDivide As New Operation("\", 2)
        Public Shared ReadOnly Modulus As New Operation("%", 2)
        Public Shared ReadOnly BitwiseAnd As New Operation("&", 2)

        Public Shared ReadOnly Power As New Operation("**", 2)
        Public Shared ReadOnly Root As New Operation("?", 2)
        Public Shared ReadOnly BitShiftLeft As New Operation("<<", 2)
        Public Shared ReadOnly BitShiftRight As New Operation(">>", 2)

        Public Shared ReadOnly PositiveSign As New Operation(" +", 1)
        Public Shared ReadOnly NegativeSign As New Operation(" -", 1)
        Public Shared ReadOnly Negate As New Operation("!", 1)

        Public Shared ReadOnly Semicolon As New Operation(";", 2)
        Public Shared ReadOnly DotDot As New Operation("..", 3)
        Public Shared ReadOnly Colon As New Operation(":", 0)

        <DebuggerHidden()> _
        Public Shared Operator =(ByVal lhs As Operation, ByVal rhs As Operation) As Boolean
            Return lhs Is rhs
        End Operator

        <DebuggerHidden()> _
        Public Shared Operator <>(ByVal lhs As Operation, ByVal rhs As Operation) As Boolean
            Return lhs IsNot rhs
        End Operator

        <DebuggerHidden()> _
        Public Shared Operator =(ByVal lhs As Operation, ByVal rhs As String) As Boolean
            Return String.Compare(lhs.Representation, rhs, True) = 0
        End Operator

        <DebuggerHidden()> _
        Public Shared Operator <>(ByVal lhs As Operation, ByVal rhs As String) As Boolean
            Return String.Compare(lhs.Representation, rhs, True) <> 0
        End Operator

        <DebuggerHidden()> _
        Public Shared Operator =(ByVal lhs As String, ByVal rhs As Operation) As Boolean
            Return String.Compare(lhs, rhs.Representation, True) = 0
        End Operator

        <DebuggerHidden()> _
        Public Shared Operator <>(ByVal lhs As String, ByVal rhs As Operation) As Boolean
            Return String.Compare(lhs, rhs.Representation, True) <> 0
        End Operator

        ''' <summary>
        ''' Parses an operator.
        ''' </summary>
        ''' <param name="op">The operator to parse.</param>
        ''' <exception cref="Expressions.ParseException">Thrown if the expression is not valid.</exception>
        ''' <returns>Returns the parsed operator.</returns>
        Public Shared Function Parse(ByVal op As String) As Operation
            Dim ret As Operation = Nothing
            If Not TryParse(op, ret) Then Throw New Expressions.ParseException(0)
            Return ret
        End Function

        ''' <summary>
        ''' Tries to parse an operator.
        ''' </summary>
        ''' <param name="op">The operator to parse.</param>
        ''' <param name="ret">The parsed operator.</param>
        ''' <returns>Returns whether parsing was successful.</returns>
        Public Shared Function TryParse(ByVal op As String, ByRef ret As Operation) As Boolean
            Dim t As Type = GetType(Operation)

            For Each m As Reflection.MemberInfo In t.FindMembers( _
                Reflection.MemberTypes.Field, _
                Reflection.BindingFlags.Public Or _
                Reflection.BindingFlags.Static, _
                Nothing, Nothing _
            )
                Dim o As Operation = DirectCast(t.InvokeMember( _
                    m.Name, _
                    Reflection.BindingFlags.GetField Or _
                    Reflection.BindingFlags.Public Or _
                    Reflection.BindingFlags.Static, _
                    Nothing, Nothing, Nothing _
                ), Operation)
                If String.Compare(op, o.Representation, True) = 0 Then
                    ret = o
                    Return True
                End If
            Next

            Return False
        End Function

#End Region

#End Region

#Region " Private "

        Private ReadOnly Representation As String

#Region " Constructors "

        ''' <remarks>As this is an enumeration type, external creation of new objects is forbidden.</remarks>
        Private Sub New(ByVal representation As String, ByVal number_of_operands As Integer)
            Me.Representation = representation
            Me.NumberOfOperands = number_of_operands
        End Sub

#End Region

#End Region

    End Class

End Namespace