Imports ParseException = madrat.Calculate.Expressions.ParseException

Namespace Calculate.Parser

    ''' <remarks>
    ''' This class implements a recursive descent parser for infix arithmetical expressions. Basically, the chosen
    ''' language is the one you'd expect by knowing conventional programming conventions for such expressions.
    ''' Therefore, the grammar is not listed in detail.
    ''' Additionally, this parser handles ranges and lists of values.
    ''' </remarks>
    Public Class InfixParser

#Region " Public "

#Region " Constructors "

        <DebuggerHidden()> _
        Public Sub New(ByVal expression As String, ByVal scanner As Scanner, ByVal stack As Expressions.Stack)
            Me.Expression = expression
            Me.Scanner = scanner
            Me.Stack = stack
        End Sub

#End Region

#Region " Properties "

        Public Property [Error]() As ParseException
            <DebuggerHidden()> _
            Get
                If _error Is Nothing Then
                    Dim et As ErrorToken = TryCast(Scanner.Lookahead, ErrorToken)

                    If et IsNot Nothing Then
                        _error = New ParseException(et.Position.Key, "Invalid symbol entered.")
                    End If
                End If

                Return _error
            End Get
            <DebuggerHidden()> _
            Private Set(ByVal value As ParseException)
                _error = value
            End Set
        End Property

#End Region

#Region " Methods "

        Public Function TryParse() As Boolean
            If Prod_Expression() AndAlso TypeOf Scanner.Lookahead Is EndOfStreamToken Then Return True

            If [Error] IsNot Nothing Then Return False

            Select Case True
                Case TypeOf Scanner.Lookahead Is ErrorToken
                    Return SetError("Invalid symbol entered.")
                Case TypeOf Scanner.Lookahead Is NumberToken, TypeOf Scanner.Lookahead Is VariableToken
                    Return SetError("Unexpected operand.")
                Case TypeOf Scanner.Lookahead Is OpenBraceToken
                    Return SetError("Unexpected opening bracket.")
                Case TypeOf Scanner.Lookahead Is CloseBraceToken
                    Return SetError("Unexpected closing bracket.")
                Case TypeOf Scanner.Lookahead Is CommaToken
                    Return SetError("Unexpected comma.")
                Case TypeOf Scanner.Lookahead Is FunctionToken
                    Return SetError("Unexpected function call.")
                Case TypeOf Scanner.Lookahead Is OperationToken
                    Return SetError("Unexpected operator.")
            End Select
        End Function

#End Region

#End Region

#Region " Private "

        Private ReadOnly Expression As String
        Private ReadOnly Scanner As Scanner
        Private ReadOnly Stack As Expressions.Stack
        Private _error As ParseException

#Region " Methods "

        Private Function Prod_Expression() As Boolean
            Return Prod_OuterExpression() AndAlso Prod_RExpression()
        End Function

        Private Function Prod_OuterExpression() As Boolean
            Return Prod_InnerExpression() AndAlso Prod_ROuterExpression()
        End Function

        Private Function Prod_InnerExpression() As Boolean
            Return Prod_Term() AndAlso Prod_RInnerExpression()
        End Function

        Private Function Prod_Term() As Boolean
            Return Prod_Factor() AndAlso Prod_RTerm()
        End Function

        Private Function Prod_Factor() As Boolean
            Return Prod_Token() AndAlso Prod_RFactor()
        End Function

        Private Function Prod_RExpression() As Boolean
            Dim la As OperationToken = TryCast(Scanner.Lookahead, OperationToken)

            If la Is Nothing Then Return True

            Select Case la.Operation
                Case Operation.Semicolon
                    Scanner.Match(la)

                    If Not Prod_OuterExpression() Then Return SetError("Missing value in field enumeration.")
                    Stack.Push(la.Operation)
                    Return Prod_RExpression()
                Case Else
                    Return True
            End Select
        End Function

        Private Function Prod_ROuterExpression() As Boolean
            Dim la As OperationToken = TryCast(Scanner.Lookahead, OperationToken)

            If la Is Nothing Then Return True

            Select Case la.Operation
                Case Operation.DotDot
                    Scanner.Match(la)

                    If Not Prod_InnerExpression() Then Return SetError("Missing term.")

                    Dim la2 As OperationToken = TryCast(Scanner.Lookahead, OperationToken)

                    If la2 IsNot Nothing AndAlso la2.Operation = Operation.Colon Then
                        Scanner.Match(la2)
                        If Not Prod_InnerExpression() Then Return SetError("Missing term.")
                    Else
                        Stack.Push(1) ' Default step value for ranges.
                    End If

                    Stack.Push(la.Operation)
            End Select

            Return True
        End Function

        Private Function Prod_RInnerExpression() As Boolean
            Dim la As OperationToken = TryCast(Scanner.Lookahead, OperationToken)

            If la Is Nothing Then Return True

            Select Case la.Operation
                Case Operation.Add, Operation.Subtract, Operation.BitwiseOr, Operation.BitwiseXor
                    Scanner.Match(la)

                    If Not Prod_Term() Then Return SetError("Missing operand.")
                    Stack.Push(la.Operation)
                    Return Prod_RInnerExpression()
                Case Else
                    Return True
            End Select
        End Function

        Private Function Prod_RTerm() As Boolean
            Dim la As OperationToken = TryCast(Scanner.Lookahead, OperationToken)

            If la Is Nothing Then Return True

            Select Case la.Operation
                Case Operation.Multipy, Operation.Divide, Operation.IntegerDivide, Operation.Modulus, Operation.BitwiseAnd
                    Scanner.Match(la)

                    If Not Prod_Factor() Then Return SetError("Missing operand.")
                    Stack.Push(la.Operation)
                    Return Prod_RTerm()
                Case Else
                    Return True
            End Select
        End Function

        Private Function Prod_RFactor() As Boolean
            Dim la As OperationToken = TryCast(Scanner.Lookahead, OperationToken)

            If la Is Nothing Then Return True

            Select Case la.Operation
                Case Operation.Power, Operation.Root, Operation.BitShiftLeft, Operation.BitShiftRight
                    Scanner.Match(la)

                    If Not Prod_Token() Then Return SetError("Missing operand.")
                    Stack.Push(la.Operation)
                    Return Prod_RFactor()
                Case Else
                    Return True
            End Select
        End Function

        Private Function Prod_Token() As Boolean
            Dim la As Token = Scanner.Lookahead

            Select Case True
                Case TypeOf la Is OpenBraceToken
                    Scanner.Match(Of OpenBraceToken)()
                    If Not Prod_Expression() Then Return SetError("Missing expression.")
                    If Not Scanner.Match(Of CloseBraceToken)() Then Return SetError("Missing closing bracket.")
                Case TypeOf la Is OperationToken
                    Dim op As Operation = DirectCast(la, OperationToken).Operation

                    If op = Operation.Add Then
                        op = Operation.PositiveSign
                    ElseIf op = Operation.Subtract Then
                        op = Operation.NegativeSign
                    End If

                    If op.NumberOfOperands <> 1 Then Return SetError("Unexpected binary or ternary operator operator.")

                    Scanner.Match(la)

                    If Not Prod_Term() Then Return SetError("Missing expression.")
                    Stack.Push(op)
                Case TypeOf la Is FunctionToken
                    Dim fun As [Function]

                    Try
                        fun = Functions.ResolveFunction(DirectCast(la, FunctionToken).Function)
                    Catch ex As Exception ' Function not found
                        Return SetError("Function '" & DirectCast(la, FunctionToken).Function & "' not found.")
                    End Try

                    Scanner.Match(Of FunctionToken)()
                    If Not Scanner.Match(Of OpenBraceToken)() Then Return SetError("Missing opening bracket.")

                    For i As Integer = 1 To fun.ArgumentNumber - 1
                        If Not Prod_Expression() Then Return SetError("Missing expression.")
                        If Not Scanner.Match(Of CommaToken)() Then Return SetError("Missing comma (probably missing parameters).")
                    Next

                    If Not Prod_Expression() Then Return SetError("Missing expression.")
                    If Not Scanner.Match(Of CloseBraceToken)() Then Return SetError("Missing closing bracket.")

                    Stack.Push(fun)
                Case TypeOf la Is NumberToken
                    Scanner.Match(Of NumberToken)()
                    Stack.Push(DirectCast(la, NumberToken).Number)
                Case TypeOf la Is VariableToken
                    Scanner.Match(Of VariableToken)()
                    Stack.Push(DirectCast(la, VariableToken).Variable)
                Case Else
                    Return False
            End Select

            Return True
        End Function

        <DebuggerHidden()> _
        Private Function SetError(ByVal message As String) As Boolean
            If [Error] Is Nothing Then [Error] = New ParseException(Scanner.Position, message)
            Return False
        End Function

#End Region

#End Region

    End Class

End Namespace