Imports RegExp = System.Text.RegularExpressions

Namespace Calculate.Parser

    ''' <summary>
    ''' Scanner base class.
    ''' </summary>
    Public MustInherit Class Scanner

#Region " Public "

#Region " Constructor "

        ''' <summary>
        ''' Creates a new instance of a scanner.
        ''' </summary>
        ''' <param name="input">The string to be scanned.</param>
        ''' <param name="start">The position in the string at which to start scanning.</param>
        <DebuggerHidden()> _
        Public Sub New(ByVal input As String, Optional ByVal start As Integer = 0)
            Me.Input = input
            Me.Start = start

            Search = New RegExp.Regex( _
                SearchPattern(), _
                RegExp.RegexOptions.Compiled Or _
                RegExp.RegexOptions.ExplicitCapture Or _
                RegExp.RegexOptions.IgnoreCase _
            )
        End Sub

#End Region

#Region " Properties "

        ''' <summary>
        ''' Returns the string to scan.
        ''' </summary>
        Public Property Input() As String
            <DebuggerHidden()> _
            Get
                Return _input
            End Get
            <DebuggerHidden()> _
            Private Set(ByVal value As String)
                _input = value
            End Set
        End Property

        ''' <summary>
        ''' Returns the starting position of the scan in the input string.
        ''' </summary>
        Public Property Start() As Integer
            <DebuggerHidden()> _
            Get
                Return _start
            End Get
            <DebuggerHidden()> _
            Private Set(ByVal value As Integer)
                _start = value
            End Set
        End Property

        ''' <summary>
        ''' Returns the next token by looking one token ahead.
        ''' </summary>
        ''' <remarks>The token returned is not consumed by the scanner.</remarks>
        Public ReadOnly Property Lookahead() As Token
            <DebuggerHidden()> _
            Get
                If LookAheadToken Is Nothing Then LookAheadToken = GetNextToken()
                Return LookAheadToken
            End Get
        End Property

        ''' <summary>
        ''' Returns the current position of the scanning process in the string.
        ''' </summary>
        Public ReadOnly Property Position() As Integer
            <DebuggerHidden()> _
            Get
                If CurrentMatch Is Nothing Then Return 0
                If Not CurrentMatch.Success Then Return Input.Length
                Return CurrentMatch.Index
            End Get
        End Property

        ''' <summary>
        ''' Returns whether the end of the input has been reached.
        ''' </summary>
        Public ReadOnly Property EndOfStream() As Boolean
            <DebuggerHidden()> _
            Get
                Return CurrentMatch IsNot Nothing AndAlso Not CurrentMatch.Success
            End Get
        End Property

#End Region

#Region " Methods "

        ''' <summary>
        ''' Matches a token type against the next token and consumes the next token.
        ''' </summary>
        ''' <typeparam name="T">The token type to match against.</typeparam>
        ''' <returns>Returns whether the token types match.</returns>
        <DebuggerHidden()> _
        Public Overloads Function Match(Of T As Token)() As Boolean
            Dim cmp As Token = GetNextToken()
            Dim dummy As Token = Lookahead
            Return TypeOf cmp Is T
        End Function

        ''' <summary>
        ''' Matches a token type against the next token and consumes the next token.
        ''' </summary>
        ''' <typeparam name="T">A token type.</typeparam>
        ''' <param name="tok">The token of which to match the type against.</param>
        ''' <returns>Returns whether the token types match.</returns>
        <DebuggerHidden()> _
        Public Overloads Function Match(Of T As Token)(ByVal tok As T) As Boolean
            Dim cmp As Token = GetNextToken()
            Dim dummy As Token = Lookahead
            Return TypeOf cmp Is T
        End Function

#End Region

#End Region

#Region " Private "

        Private _input As String
        Private _start As Integer
        Private _current_match As RegExp.Match
        Private _look_ahead_token As Token
        Private _token_types As Dictionary(Of String, Token)

        Protected ReadOnly Search As RegExp.Regex

#Region " Properties "

        Private Property CurrentMatch() As RegExp.Match
            <DebuggerHidden()> _
            Get
                Return _current_match
            End Get
            <DebuggerHidden()> _
            Set(ByVal value As RegExp.Match)
                _current_match = value
            End Set
        End Property

        Private Property LookAheadToken() As Token
            <DebuggerHidden()> _
            Get
                Return _look_ahead_token
            End Get
            <DebuggerHidden()> _
            Set(ByVal value As Token)
                _look_ahead_token = value
            End Set
        End Property

#End Region

#Region " Methods "

        ''' <summary>
        ''' Sets the regular expression search pattern to be used by the scanner.
        ''' </summary>
        ''' <returns>Returns the search pattern.</returns>
        Protected MustOverride Function SearchPattern() As String

        ''' <summary>
        ''' Callback method reacting individually on the current match and returning
        ''' a corresponding token, or nothing if the token should be ignored (in which
        ''' case the next token will be read).
        ''' </summary>
        ''' <param name="match">The current match in the search string.</param>
        ''' <returns>Returns a corresponding token.</returns>
        Protected MustOverride Function GetTokenFromMatch(ByVal match As RegExp.Match) As Token

        Private Function GetNextToken() As Token
            If LookAheadToken IsNot Nothing Then
                Dim ret As Token = LookAheadToken
                LookAheadToken = Nothing
                Return ret
            End If

            If CurrentMatch Is Nothing Then
                CurrentMatch = Search.Match(Input, Start)
            Else
                CurrentMatch = CurrentMatch.NextMatch()
            End If

            If EndOfStream Then Return New EndOfStreamToken()

            GetNextToken = GetTokenFromMatch(CurrentMatch)

            ' Token should be omitted.
            If GetNextToken Is Nothing Then Return GetNextToken()
        End Function

#End Region

#End Region

    End Class

End Namespace